/*                                                                           
**  fi_lib++  --- A fast interval library (Version 2.0)                     
**                                                                  
**  Copyright (C) 2001:                                                        
**                                                     
**  Werner Hofschuster, Walter Kraemer                               
**  Wissenschaftliches Rechnen/Softwaretechnologie (WRSWT)  
**  Universitaet Wuppertal, Germany                                           
**  Michael Lerch, German Tischler, Juergen Wolff von Gudenberg       
**  Institut fuer Informatik                                         
**  Universitaet Wuerzburg, Germany                                           
** 
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/
#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <iostream>
#include <cfloat>

#include <sys/types.h>

#if defined(__linux) && defined(__i386__)
typedef unsigned short uint16_t;
#endif

#if defined(__i386__)
int main()
{
	double a = DBL_MAX, b = 1.0, c = 0.0, d = 0.0;
	uint16_t const cw_down = 0x067F;

	asm 
	(
		"fldcw %3; fldl %1 ; fchs ; fldl %2 ; fchs ; faddp ; fchs; fstpl %0"
		: "=m" (c) : "m" (a), "m" (b), "m" (cw_down) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
	);

	asm 
	(
		"fldcw %3; fldl %1 ; fchs ; fldl %2 ; fchs ; faddp ; fstpl %0; fldl %0; fchs; fstpl %0"
		: "=m" (d) : "m" (a), "m" (b), "m" (cw_down) :
		"st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
	);


	if ( c != d )
	{
		std::cout << "fldcw cw_down; fldl DBL_MAX ; fchs ; fldl 1.0 ; fchs ; faddp ; fchs ; fstpl c" << std::endl;
		std::cout << "Unreloaded\t- c = nabla(-DBL_MAX+-1.0) = " << c << std::endl;
		std::cout << "fldcw cw_down; fldl DBL_MAX ; fchs ; fldl 1.0 ; fchs ; faddp ; fstpl c; fldl c; fchs; fstpl c" << std::endl;
		std::cout << "Reloaded  \t- c = nabla(-DBL_MAX+-1.0) = " << d << std::endl;
		std::cout << "Unreloaded result and reloaded result do not match." << std::endl;
		std::cout << "Either your compiler/assembler or processor is buggy !" << std::endl;
	}
	else
		std::cout << "Test succeeded." << std::endl;
}
#else
int main()
{
	std::cout << "No test for your archtecture here." << std::endl;
}
#endif
