/*                                                                           
**  fi_lib++  --- A fast interval library (Version 2.0)                     
**                                                                  
**  Copyright (C) 2001:                                                        
**                                                     
**  Werner Hofschuster, Walter Kraemer                               
**  Wissenschaftliches Rechnen/Softwaretechnologie (WRSWT)  
**  Universitaet Wuppertal, Germany                                           
**  Michael Lerch, German Tischler, Juergen Wolff von Gudenberg       
**  Institut fuer Informatik                                         
**  Universitaet Wuerzburg, Germany                                           
** 
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/
#include <interval/interval.hpp>

#include <cstdlib>
#include <ctime>

double rd()
{
	double number;
	unsigned int sign,expo,mant0,mant1;

	do
	{
		sign = std::rand()%2;
		expo = std::rand()%0x800;
		mant1 = std::rand() % (21 << 1);
		mant0 = std::rand();

		number = filib::primitive::compose(sign,expo,mant0,mant1);
	} while ( expo == 0x7FF && (mant0 || mant1) );

	return number;
}

void pd(double const & n, std::string const & de, std::ostream & out = std::cerr)
{
	unsigned int a0,a1,a2,a3;

	filib::primitive::decompose(n,a0,a1,a2,a3);

	out << de << "[" << a0 << "," << a1 << "," << a2 << "," << a3 << "]";
}

int main()
{
	filib::fp_traits<double,filib::native_switched>::setup();

	srand(std::time(0));

	for ( size_t j = 5; j; --j )
	{
		for ( size_t i = 0; i < 1000000; ++i )
		{
			double inf0 = rd(), sup0 = rd();
			double inf1 = rd(), sup1 = rd();

			if ( inf0 > sup0 )
				std::swap(inf0,sup0);
			if ( inf1 > sup1 )
				std::swap(inf1,sup1);

			filib::interval<double,filib::native_switched,filib::i_mode_extended> 
				ra(inf0,sup0), rb(inf1,sup1);
			filib::interval<double,SUBST_MODE,filib::i_mode_extended> 
				a(inf0,sup0), b(inf1,sup1);
	
			filib::interval<double,filib::native_switched,filib::i_mode_extended>
				rc = ra+rb;
			filib::interval<double,SUBST_MODE,filib::i_mode_extended>
				c = a+b;
	
			if ( c.inf() > rc.inf() )
			{
				std::cerr << "inf failure + " << i << std::endl;

				pd (a.inf(), "a.inf: ");
				std::cerr<< std::endl;
				pd (ra.inf(), "ra.inf: ");
				std::cerr << std::endl;

				pd (b.inf(), "b.inf: ");
				std::cerr<< std::endl;
				pd (rb.inf(), "rb.inf: ");
				std::cerr << std::endl;

				pd (c.inf(), "c.inf: ");
				std::cerr<< std::endl;
				pd (rc.inf(), "rc.inf: ");
				std::cerr << std::endl;
			}
			if ( c.sup() < rc.sup() )
			{
				std::cerr << "sup failure + " << i << std::endl;

				pd (a.sup(), "a.sup: ");
				std::cerr<< std::endl;
				pd (ra.sup(), "ra.sup: ");
				std::cerr << std::endl;

				pd (b.sup(), "b.sup: ");
				std::cerr<< std::endl;
				pd (rb.sup(), "rb.sup: ");
				std::cerr << std::endl;

				pd (c.sup(), "c.sup: ");
				std::cerr<< std::endl;
				pd (rc.sup(), "rc.sup: ");
				std::cerr << std::endl;
			}
		}
		for ( size_t i = 0; i < 1000000; ++i )
		{
			double inf0 = rd(), sup0 = rd();
			double inf1 = rd(), sup1 = rd();

			if ( inf0 > sup0 )
				std::swap(inf0,sup0);
			if ( inf1 > sup1 )
				std::swap(inf1,sup1);

			filib::interval<double,filib::native_switched,filib::i_mode_extended> 
				ra(inf0,sup0), rb(inf1,sup1);
			filib::interval<double,SUBST_MODE,filib::i_mode_extended> 
				a(inf0,sup0), b(inf1,sup1);
	
			filib::interval<double,filib::native_switched,filib::i_mode_extended>
				rc = ra-rb;
			filib::interval<double,SUBST_MODE,filib::i_mode_extended>
				c = a-b;
	
			if ( c.inf() > rc.inf() )
			{
				std::cerr << "inf failure - " << i << std::endl;

				pd (a.inf(), "a.inf: ");
				std::cerr<< std::endl;
				pd (ra.inf(), "ra.inf: ");
				std::cerr << std::endl;

				pd (b.inf(), "b.inf: ");
				std::cerr<< std::endl;
				pd (rb.inf(), "rb.inf: ");
				std::cerr << std::endl;

				pd (c.inf(), "c.inf: ");
				std::cerr<< std::endl;
				pd (rc.inf(), "rc.inf: ");
				std::cerr << std::endl;
			}
			if ( c.sup() < rc.sup() )
			{
				std::cerr << "sup failure - " << i << std::endl;

				pd (a.sup(), "a.sup: ");
				std::cerr<< std::endl;
				pd (ra.sup(), "ra.sup: ");
				std::cerr << std::endl;

				pd (b.sup(), "b.sup: ");
				std::cerr<< std::endl;
				pd (rb.sup(), "rb.sup: ");
				std::cerr << std::endl;

				pd (c.sup(), "c.sup: ");
				std::cerr<< std::endl;
				pd (rc.sup(), "rc.sup: ");
				std::cerr << std::endl;
			}
		}
		for ( size_t i = 0; i < 1000000; ++i )
		{
			double inf0 = rd(), sup0 = rd();
			double inf1 = rd(), sup1 = rd();

			if ( inf0 > sup0 )
				std::swap(inf0,sup0);
			if ( inf1 > sup1 )
				std::swap(inf1,sup1);

			filib::interval<double,filib::native_switched,filib::i_mode_extended> 
				ra(inf0,sup0), rb(inf1,sup1);
			filib::interval<double,SUBST_MODE,filib::i_mode_extended> 
				a(inf0,sup0), b(inf1,sup1);
	
			filib::interval<double,filib::native_switched,filib::i_mode_extended>
				rc = ra*rb;
			filib::interval<double,SUBST_MODE,filib::i_mode_extended>
				c = a*b;
	
			if ( c.inf() > rc.inf() )
			{
				std::cerr << "inf failure * " << i << std::endl;

				pd (a.inf(), "a.inf: ");
				std::cerr<< std::endl;
				pd (ra.inf(), "ra.inf: ");
				std::cerr << std::endl;

				pd (b.inf(), "b.inf: ");
				std::cerr<< std::endl;
				pd (rb.inf(), "rb.inf: ");
				std::cerr << std::endl;

				pd (c.inf(), "c.inf: ");
				std::cerr<< std::endl;
				pd (rc.inf(), "rc.inf: ");
				std::cerr << std::endl;
			}
			if ( c.sup() < rc.sup() )
			{
				std::cerr << "sup failure * " << i << std::endl;

				pd (a.sup(), "a.sup: ");
				std::cerr<< std::endl;
				pd (ra.sup(), "ra.sup: ");
				std::cerr << std::endl;

				pd (b.sup(), "b.sup: ");
				std::cerr<< std::endl;
				pd (rb.sup(), "rb.sup: ");
				std::cerr << std::endl;

				pd (c.sup(), "c.sup: ");
				std::cerr<< std::endl;
				pd (rc.sup(), "rc.sup: ");
				std::cerr << std::endl;
			}
		}
		for ( size_t i = 0; i < 1000000; ++i )
		{
			double inf0 = rd(), sup0 = rd();
			double inf1 = rd(), sup1 = rd();

			if ( inf0 > sup0 )
				std::swap(inf0,sup0);
			if ( inf1 > sup1 )
				std::swap(inf1,sup1);

			filib::interval<double,filib::native_switched,filib::i_mode_extended> 
				ra(inf0,sup0), rb(inf1,sup1);
			filib::interval<double,SUBST_MODE,filib::i_mode_extended> 
				a(inf0,sup0), b(inf1,sup1);
	
			filib::interval<double,filib::native_switched,filib::i_mode_extended>
				rc = ra/rb;
				filib::interval<double,SUBST_MODE,filib::i_mode_extended>
				c = a/b;
	
			if ( c.inf() > rc.inf() )
			{
				std::cerr << "inf failure / " << i << std::endl;

				pd (a.inf(), "a.inf: ");
				std::cerr<< std::endl;
				pd (ra.inf(), "ra.inf: ");
				std::cerr << std::endl;

				pd (b.inf(), "b.inf: ");
				std::cerr<< std::endl;
				pd (rb.inf(), "rb.inf: ");
				std::cerr << std::endl;

				pd (c.inf(), "c.inf: ");
				std::cerr<< std::endl;
				pd (rc.inf(), "rc.inf: ");
				std::cerr << std::endl;
			}
			if ( c.sup() < rc.sup() )
			{
				std::cerr << "sup failure / " << i << std::endl;

				pd (a.sup(), "a.sup: ");
				std::cerr<< std::endl;
				pd (ra.sup(), "ra.sup: ");
				std::cerr << std::endl;

				pd (b.sup(), "b.sup: ");
				std::cerr<< std::endl;
				pd (rb.sup(), "rb.sup: ");
				std::cerr << std::endl;

				pd (c.sup(), "c.sup: ");
				std::cerr<< std::endl;
				pd (rc.sup(), "rc.sup: ");
				std::cerr << std::endl;
			}
		}
		std::cout << "." << std::flush;
	}
	std::cout << std::endl;
}
