/*                                                                           
**  fi_lib++  --- A fast interval library (Version 2.0)                     
**                                                                  
**  Copyright (C) 2001:                                                        
**                                                     
**  Werner Hofschuster, Walter Kraemer                               
**  Wissenschaftliches Rechnen/Softwaretechnologie (WRSWT)  
**  Universitaet Wuppertal, Germany                                           
**  Michael Lerch, German Tischler, Juergen Wolff von Gudenberg       
**  Institut fuer Informatik                                         
**  Universitaet Wuerzburg, Germany                                           
** 
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*/
#ifndef _FILIB_ROUNDING_H
#define _FILIB_ROUNDING_H

#include "test/config.hpp"

#if defined(FILIB_NATIVE_ROUNDING)

 #if defined(__i386__)

   // 387 control word for directed roundings; no interrupts and double
   // precision rounding 
   #define CW_ROUND_NEAR 0x027F
   #define CW_ROUND_DOWN 0x067F
   #define CW_ROUND_UP   0x0A7F

   

   #if defined (__GNUC__)

     // Macros for switching rounding mode 
     #define ROUND_DOWN\
     asm("fldcw %0" : : "m" (CW_ROUND_DOWN) : "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)")

     #define ROUND_UP\
     asm("fldcw %0" : : "m" (CW_ROUND_UP) : "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)")

     #define ROUND_NEAR\
     asm("fldcw %0" : : "m" (CW_ROUND_NEAR) : "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)")

     // Macros for updating arithmetic operations with directed roundings
     // x op= y
     #define ADD_DOWN_UPD(x, y) { \
       ROUND_DOWN; x += y; \
     }

     #define ADD_UP_UPD(x, y) { \
       ROUND_UP; x += y; \
     }

     #define SUB_DOWN_UPD(x, y) { \
       ROUND_DOWN; x -= y; \
     }

     #define SUB_UP_UPD(x, y) { \
       ROUND_UP; x -= y; \
     }

     #define MUL_DOWN_UPD(x, y) { \
       ROUND_DOWN; x *= y; \
     }

     #define MUL_UP_UPD(x, y) { \
       ROUND_UP; x *= y; \
     }

     #define DIV_DOWN_UPD(x, y) { \
       ROUND_DOWN; x /= y; \
     }

     #define DIV_UP_UPD(x, y) { \
       ROUND_UP; x /= y; \
     }

     // Macros for arithmetic operations with directed roundings
     // z = x op y  
     #define ADD_DOWN(z, x, y) { \
       ROUND_DOWN; z = x + y; \
     } 

     #define ADD_UP(z, x, y) { \
       ROUND_UP; z = x + y; \
     } 

     #define SUB_DOWN(z, x, y) { \
       ROUND_DOWN; z = x - y; \
     } 

     #define SUB_UP(z, x, y) { \
       ROUND_UP; z = x - y; \
     } 

     #define MUL_DOWN(z, x, y) { \
       ROUND_DOWN; z = x * y; \
     } 

     #define MUL_UP(z, x, y) { \
       ROUND_UP; z = x * y; \
     } 

     #define DIV_DOWN(z, x, y) { \
       ROUND_DOWN; z = x / y; \
     } 

     #define DIV_UP(z, x, y) { \
       ROUND_UP; z = x / y; \
     }   

   #elif defined (__KCC)

     extern "C" void * __kai_apply(const char *,...);

     // Macros for switching rounding mode 
     #define ROUND_NEAR\
     __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_NEAR);


     // Macros for updating arithmetic operations with directed roundings
     // x op= y
     #define ADD_DOWN_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) += (%e);", &(x), (y)); \
     }

     #define ADD_UP_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) += (%e);", &(x), (y)); \
     }

     #define SUB_DOWN_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) -= (%e);", &(x), (y)); \
     }

     #define SUB_UP_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) -= (%e);", &(x), (y)); \
     }

     #define MUL_DOWN_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) *= (%e);", &(x), (y)); \
     }

     #define MUL_UP_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) *= (%e);", &(x), (y)); \
     }

     #define DIV_DOWN_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) /= (%e);", &(x), (y)); \
     }

     #define DIV_UP_UPD(x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) /= (%e);", &(x), (y)); \
     }


     // Macros for arithmetic operations with directed roundings
     // z = x op y 
     #define ADD_DOWN(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) = (%e) + (%e);", &(z), (x), (y)); \
     }

     #define ADD_UP(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) = (%e) + (%e);", &(z), (x), (y)); \
     }

     #define SUB_DOWN(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) = (%e) - (%e);", &(z), (x), (y)); \
     }

     #define SUB_UP(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) = (%e) - (%e);", &(z), (x), (y)); \
     }

     #define MUL_DOWN(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) = (%e) * (%e);", &(z), (x), (y)); \
     }

     #define MUL_UP(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) = (%e) * (%e);", &(z), (x), (y)); \
     }

     #define DIV_DOWN(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_DOWN); \
       __kai_apply("(%a) = (%e) / (%e);", &(z), (x), (y)); \
     }

     #define DIV_UP(z, x, y) { \
       __kai_apply("__asm__(\"fldcw %%0\" : : \"m\" (%e));", CW_ROUND_UP); \
       __kai_apply("(%a) = (%e) / (%e);", &(z), (x), (y)); \
     }

   #else 
   
     #error Compiler not supported for directed rounding !

   #endif 

 #else 

   #error Architecture not supported for directed rounding !

 #endif // __i386__


 #ifdef FILIB_ENSURE_ROUND_NEAR
   #define POSS_ROUND_NEAR ROUND_NEAR
 #else
   #define POSS_ROUND_NEAR 
 #endif

#endif // FILIB_NATIVE_ROUNDING


#endif
